%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Сайт:     www.rotr.info
%% Автор:    hamper
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%   Численное решение задачи расчёта вихревых токов   %%%%%%%%%%%%%

% Рассматривается частный случай: вихревые токи в проводящем цилиндре,
% находящемся во внешнем переменном магнитном поле.
% Предполагается, что магнитное поле обладает осевой симметрией относительно
% оси цилиндра.
% Предполагается, что магнитная проницаемость среды внутри и вне цилиндра
% одинакова.

% Магнитное поле описывается с помощью векторного потенциала A. 
% В рассматриваемом случае только одна его проекция (азимутальная) отлична от 0,
% поэтому всюду подразумевается использование этой проекции.

% Полное магнитное поле имеет две составляющие: внешне поле (поле индуктора или
% катушки) и магнитное поле, создаваемое вихревыми токами, возникающими в
% проводящем цилиндре:
% A=A0+Ae.

% Внешнее поле считается заданным (если это поле, создаваемое круглым витком, 
% расположенным соосно с цилиндром или катушки с малым поперечным сечением, оно
% может быть рассчитано с помощью функции A0field).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%   Некоторые постоянные   %%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Магнитная постоянная.
mu0=1.25663706212e-6;

% Удельная проводимость некоторых материалов.
sigma_cu=56e6;		% медь
sigma_ag=62.9e6;	% серебро
sigma_au=43.5e6;	% золото
sigma_brass=13e6;	% латунь

% Удельное сопротивление грунта обычно находится в пределах 20..100 Ом*м,
% если в качестве оценочного значения будем использовать 100, то
% удельная проводимость
sigma_gnd=0.01;

% Расстояния меньше этого будем считать равными 0
% (там, где нулевые расстояния требуют отдельного варианта вычислений).
global lmin=1e-9;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%   Функции   %%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Расчёт поля в точке.
% Расчёт магнитного поля кругового линейного тока с радиусом контура a при силе
% тока в контуре 1А. При иной величине силы тока, полученный результат нужно
% умножить на силу тока, в действительности протекающего по контуру.
% Функция вычисляет проекцию theta векторного потенциала в точке 
% с координатами (r, z), заданными в цилиндрической системе координат
% с началом в центре контура с током и осью Oz, совпадающей с осью контура. 
% (Только эта проекция в данном случае может быть отлична от нуля, причём она
% не зависит от угла theta).
% Параметры функции.
% a - радиус контура с током, сила тока принимается равной 1А;
% r, z - координаты точки, для которой вычисляется theta-проекция векторного
% потенциала;
% mu_abs - абсолютная магнитная проницаемость среды, заполняющей всё
% пространство, окружающее контур, mu_abs=mu*mu0.

function y=A0point(a, r, z, mu_abs)
  global lmin
  
  if r<lmin
    y=0; return;
  endif
  
  bq=4*a*r./(r.^2+z.^2+a.^2+2*a*r);   % здесь bq - параметр, известный как b^2
  % В Octave, функции для вычисления эллиптических интегралов передаётся параметр
  % b^2; в других системах аналогичная функция может требовать параметр b
  % (уточняйте в документации).
  [K, E]=ellipke(bq);
  y=mu_abs*sqrt(a./r./bq)./(2*pi).*((2-bq).*K-2*E);
endfunction

% Расчёт поля в области (в узлах сетки, заданной для полусечения цилиндра).
% Рассчитать магнитное поле, создаваемое круговым витком с током 1А
% в цилиндрической области, соосной с витком.
% Векторный потенциал A0 в цилиндрической системе координат с началом в центре
% витка и осью Oz, совпадающей с осью витка, имеет единственную ненулевую 
% проекцию theta, причём эта проекция не зависит от угла theta.
% Поэтому, чтобы полностью описать поле в цилиндрической области,
% достаточно выполнить расчёт для точек половины какого-либо сечения
% цилиндра (сечения, проходящего через ось Oz), например, для точек
% с координатами theta=0, r=0..R, z=z0..z0-H.
% Данную область равномерно разбиваем на nrows*ncols узлов, задающих
% прямоугольную сетку и результат возвращаем в виде матрицы, содержащей значения
% A0 для узлов сетки (увеличение номера строки в матрице соответствует движению
% вниз; увеличение номера столбца соответствует 
% увеличению координаты r соответствующего узла).
% Параметры функции.
% a1 - радиус контура (витка) с током;
% z0 - z-координата верхнего торца цилиндрической области, в которой
% рассчитываем поле (z0<0, если цилиндр находится ниже контура);
% H - высота цилиндрической области;
% R - радиус цилиндрической области;
% nrows, ncols - количество строк и столбцов в матрице-результате;
% mu_abs - абсолютная магнитная проницаемость среды (предполагается всюду
% одинаковой, и в области с вихревыми токами, и вне неё).
% Результат - матрица, элемент с индексами (p, q) которой содержит величину
% A0 для узла (p, q) прямоугольной сетки, на которую разбито рассматриваемое
% сечение цилиндра.

function y=A0field(a1, z0, H, R, nrows, ncols, mu_abs)
	y=zeros(nrows, ncols);
	h2=H/(nrows-1);
	h1=R/(ncols-1);

	for p=1:nrows
		for q=2:ncols
			y(p, q)=A0point(a1, h1*(q-1), z0-h2*(p-1), mu_abs);
		endfor
	endfor
endfunction

% Найти в заданной точке пространства магнитное поле, создаваемое вихревыми
% токами, текущими в проводящем цилиндре, если известно полное магнитное поле
% в области цилиндра (предполагается, что магнитное поле обладает осевой
% симметрией относительно оси цилиндра).
% 
% Цилиндр с радиусом R и высотой H находится в переменном магнитном поле A
% (которое складывается из внешнего поля A0 и поля вихревых токов Ae).
% Поле обладает осевой симметрией относительно оси цилиндра
% (имеет единственную ненулевую проекцию Atheta, которая не зависит от theta).
% Поле (амплитудные значения) задано матрицей A таким образом, как это указано 
% в описании к функции A0field (т.е. в узлах сетки).
% Функция вычисляет поле в точке с заданными координатами (точка может
% быть как вне цилиндра, так и внутри - в узлах сетки), создаваемое вихревыми
% токами в данной цилиндрической области.
% Аргументы функции.
% A - матрица, задающая полное магнитное поле в цилиндрической области,
% заполненной проводящим материалом;
% w - циклическая частота колебаний поля (w=2*pi*f);
% sigma - удельная проводимость среды в цилиндре с вихревыми токами;
% H - высота цилиндрической области с вихревыми токами;
% R - радиус цилиндрической области;
% h - высота точки, в которой вычисляется поле над верхним торцом цилиндрической
% области (h<0, если точка находится ниже верхнего торца);
% r - r-координата точки, в которой вычисляется поле;
% mu_abs - абсолютная магнитная проницаемость среды (предполагается всюду
% одинаковой, и в области с вихревыми токами, и вне неё).

function y=Aeddy(A, w, sigma, H, R, h, r, mu_abs)
  global lmin
  
  if r<lmin 
    y=0; return;
  endif
  
  [nrows, ncols]=size(A);
  h1=R/(ncols-1);
  h2=H/(nrows-1);
  S=h1*h2;
  c=-0.25*i*w*sigma*S;
  
  y=0;
  
  for p=1:nrows-1
    for q=1:ncols-1
      k=A0point(h1*(q-0.5), r, h+h2*(p-0.5), mu_abs);
      y+=c*k*(A(p, q)+A(p, q+1)+A(p+1, q)+A(p+1, q+1));
    endfor
  endfor
endfunction

% Служебная функция для перехода от индексации с двумя индексами к одноиндексной.
% Преобразует индекс элемента матрицы в условный порядковый номер элемента;
% p, q - номер строки и номер столбца, где находится элемент;
% ncols - количество столбцов в матрице.

function y=flatnum(p, q, ncols)
  y=(p-1)*ncols+q;
endfunction

% Служебная функция для вычисления вектора весовых коэффициентов,
% связывающих полное магнитное поле в проводящем цилиндре и магнитное
% поле только вихревых токов в заданной точке
% (весовые коэффициенты определяют вклад полного векторного потенциала
% в каждой точке проводящего цилиндра в векторный потенциал только
% вихревых токов, вычисляемый в заданной точке).
% Используется связь вектора плотности тока и векторного потенциала, для
% комплексных амплитуд проекций величин в нашем случае соотношение имеет вид
% (в некоторой точке): j=-i*w*sigma*A.
% Предполагается переход к линейной нумерации точек рассматриваемой области.
% Для вычисления используется следующее приближение.
% Распределённый ток каждой ячейки, заключённой между 4 узлами сетки заменяется 
% линейным током, проходящим через центр ячейки, в который каждый из узлов
% вносит вклад 1/4*j*S, где S - площадь ячейки.
% Аргументы функции.
% nrows, ncols - количество строк и столбцов в матрице, задающей поле в
% цилиндрической области с вихревыми токами
% (о матрице смотрите в описании функции A0field);
% w - циклическая частота переменного магнитного поля;
% sigma - удельная проводимость среды в цилиндре с вихревыми токами;
% H - высота цилиндрической области с вихревыми токами;
% R - радиус этой области;
% h - высота точки относительно верхнего торца цилиндрической области (h<0, если
% точка находится ниже верхнего торца), для которой вычисляются коэффициенты;
% r - r-координата этой точки;
% mu_abs - абсолютная магнитная проницаемость среды (предполагается всюду
% одинаковой, и в области с вихревыми токами, и вне неё).

function y=calc_a_weights(nrows, ncols, w, sigma, H, R, h, r, mu_abs)
  n=nrows*ncols;
  y=zeros(1, n);
  
  h1=R/(ncols-1);
  h2=H/(nrows-1);
  S=h1*h2;
  c=-0.25*i*w*sigma*S;
  
  for p=1:nrows-1
    for q=1:ncols-1
      k=A0point(h1*(q-0.5), r, h+h2*(p-0.5), mu_abs);
      y(flatnum(p, q, ncols))+=k;
      y(flatnum(p, q+1, ncols))+=k;
      y(flatnum(p+1, q, ncols))+=k;
      y(flatnum(p+1, q+1, ncols))+=k;
    endfor
  endfor
  
  y*=c;
endfunction

% Служебная функция.
% По карте распределения векторного потенциала и вектору весовых коэффициентов,
% определить векторный потенциал в точке, для которой был составлен этот
% вектор весовых коэффициентов.

function y=calcAonW(Ax, W)
  [nrows, ncols]=size(Ax);
  n=nrows*ncols;
  if n!=size(W)
    error("Несоответствие размеров.");
  endif
  y=0;
  for p=1:nrows
    for q=1:ncols
      y+=Ax(p, q)*W(flatnum(p, q, ncols));
    endfor
  endfor
endfunction

% Служебная функция.
% Составить систему линейных уравнений для приближённого решения задачи
% по вычислению составляющей магнитного поля в проводящем цилиндре,
% создаваемой только вихревыми токами в этом цилиндре, возникающими под
% действием заданного внешнего переменного магнитного поля A0, обладающего
% осевой симметрией относительно оси цилиндра
% (вычислить коэффициенты этой системы линейных уравнений).
% Параметры функции.
% A0 - матрица, задающая поле индуктора (смотрите описание к функции A0field);
% w - циклическая частота переменного магнитного поля;
% sigma - удельная проводимость среды, заполняющей цилиндрическую область с
% вихревыми токами;
% H - высота цилиндрической области с вихревыми токами;
% R - радиус этой области;
% mu_abs - абсолютная магнитная проницаемость среды (предполагается всюду
% одинаковой, и в области с вихревыми токами, и вне неё).

function [M, b]=create_pde(A0, w, sigma, H, R, mu_abs)
  [nrows ncols]=size(A0);
  
  h2=H/(nrows-1);
  h1=R/(ncols-1);
  
  n=nrows*ncols;
  
  M=zeros(n);
  b=zeros(n, 1);
  
  neq=1;  % текущий номер уравнения
  
  % ДУ для внутренних точек области.
  for p=2:nrows-1
    for q=2:ncols-1
      M(neq, flatnum(p, q+1, ncols))=q/((q-1)*h1^2);
      M(neq, flatnum(p, q, ncols))=-2/h1^2-1/(q-1)/h1^2-2/h2^2-1/((q-1)^2*h1^2)-i*w*sigma*mu_abs;
      M(neq, flatnum(p, q-1, ncols))=1/h1^2;
      M(neq, flatnum(p+1, q, ncols))=1/h2^2;
      M(neq, flatnum(p-1, q, ncols))=1/h2^2;
      b(neq)=i*w*sigma*mu_abs*A0(p, q);
      neq++;
    endfor
  endfor
  
  % граничные условия на оси (A=0)
  for p=1:nrows
    M(neq, flatnum(p, 1, ncols))=1;
    neq++;
  endfor
  
  % индексы оставшихся граничных точек
  bp_count=nrows+2*ncols-4;
  bps=zeros(2, bp_count);
  bpi=1;
  for p=2:ncols-1
    bps(1, bpi)=1;
    bps(2, bpi)=p;
    bpi++;
    bps(1, bpi)=nrows;
    bps(2, bpi)=p;
    bpi++;
  endfor
  for p=1:nrows
    bps(1, bpi)=p;
    bps(2, bpi)=ncols;
    bpi++;
  endfor
  if bpi!=bp_count+1
    error("Internal error");
  endif
  
  % Интегральные условия для каждой из указанных граничных точек.
  for k=1:bp_count
    p=bps(1, k);
    q=bps(2, k);
        
    W=calc_a_weights(nrows, ncols, w, sigma, H, R, -h2*(p-1), h1*(q-1), mu_abs);
    M(neq, :)=W;
    M(neq, flatnum(p, q, ncols))-=1;
    
    bb=0;
    for pp=1:nrows
      for qq=1:ncols
        bb-=A0(pp, qq)*W(flatnum(pp, qq, ncols));
      endfor
    endfor
    b(neq)=bb;
    neq++;
  endfor

  if neq!=n+1
    error("internal error: wrong number of equations.");
  endif
endfunction

% Составить систему линейных уравнений для решения задачи о расчёте поля
% вихревых токов и решить эту систему.
% Параметры функции.
% A0 - матрица, задающая поле индуктора (смотрите описание к функции A0field);
% w - циклическая частота переменного магнитного поля;
% sigma - удельная проводимость среды, заполняющей цилиндрическую область с
% вихревыми токами;
% H - высота цилиндрической области с вихревыми токами;
% R - радиус этой области;
% mu_abs - абсолютная магнитная проницаемость среды (предполагается всюду
% одинаковой, и в области с вихревыми токами, и вне неё).

function y=solve_pde(A0, w, sigma, H, R, mu_abs)
  [M b]=create_pde(A0, w, sigma, H, R, mu_abs);
  ylin=M\b;
  [nrows ncols]=size(A0);
  y=zeros(nrows, ncols);
  for p=1:nrows
    for q=1:ncols
      y(p, q)=ylin(flatnum(p, q, ncols));
    endfor
  endfor
endfunction

% Вычислить изменение комплексного сопротивления контура в результате влияния
% вихревых токов.
% Параметры функции.
% a1 - радиус контура (индуктора поля);
% a2 - радиус цилиндра (соосного с контуром) с вихревыми токами;
% d2 - высота цилиндра;
% h - расстояние между плоскостью контура (индуктора) и плоскостью ближайшего 
% торца цилиндра; если цилиндр пересекает плоскость катушки, в качестве
% аргумента используем значение -h, где h - расстояние от плоскости катушки до
% любого торца цилиндра (передаём функции отрицательное значение);
% w - циклическая частота переменного тока в контуре;
% sigma - удельная проводимость материала цилиндра;
% mu_abs - абсолютная магнитная проницаемость среды;
% nrows, ncols определяют количество ячеек сетки, используемой при решении задачи.

function y=solve_md(a1, a2, d2, h, w, sigma, mu_abs, nrows, ncols)
	A0=A0field(a1, -h, d2, a2, nrows, ncols, mu_abs);
	Ae=solve_pde(A0, w, sigma, d2, a2, mu_abs);
	y=2*i*w*pi*a1*Aeddy(A0+Ae, w, sigma, d2, a2, h, a1, mu_abs);
endfunction

% Оценить относительное изменение индуктивности
% (предполагается, что индуктивность катушки примерно равна L1=4.242*mu0*n1^2*a1).
% В отличие от результата функции solve_md, относительное изменение не зависит
% от количества витков n1 в катушке.
% Параметры функции.
% a1 - радиус контура (индуктора поля);
% a2 - радиус цилиндра (соосного с контуром) с вихревыми токами;
% d2 - высота цилиндра;
% h - расстояние между плоскостью контура (индуктора) и плоскостью ближайшего 
% торца цилиндра или -h, если цилиндр пересекает плоскость катушки (в этом
% случае h - расстояние до любого торца цилиндра);
% w - циклическая частота переменного тока в контуре;
% sigma - удельная проводимость материала цилиндра;
% mu_abs - абсолютная магнитная проницаемость среды (предполагается всюду
% одинаковой, и в области с вихревыми токами, и вне неё);
% nrows, ncols определяют количество ячеек сетки, используемой при решении задачи.

function y=calc_el1(a1, a2, d2, h, w, sigma, mu_abs, nrows, ncols)
	dz1=solve_md(a1, a2, d2, h, w, sigma, mu_abs, nrows, ncols);
	y=imag(dz1)/w/(4.242*mu_abs*a1);
endfunction


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%   Пример использования   %%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% 1. Задаём геометрические параметры задачи, свойства среды и частоту сигнала.
%% Здесь: анализ поля в латунной монете 5коп. (диаметр 25мм, толщина 1.5мм),
%% находящейся на расстоянии 50мм от рамки.
%% Радиус индуктора 100мм.
%% Используется сетка 25*25 узлов.
%% Частота 5000Гц.
  % a1=0.1; z0=-0.05; H=1.5e-3; R=25e-3 / 2; nrows=25; ncols=25;
  % w=2*pi*5000; sigma=sigma_brass;

%% Рассчитываем составляющую поля, создаваемую катушкой в области исследования.
  % A0=A0field(a1, z0, H, R, nrows, ncols, mu0);

%% Решаем задачу по вычислению магнитного поля вихревых токов. 
  % Ae=solve_pde(A0, w, sigma, H, R, mu0);

%% Магнитное поле равно сумме двух составляющих.
  % A=A0+Ae;

%% Выводим карту распределения поля.
  % pcolor(abs(A));
  
%% Вычисляем магнитное поле, создаваемое вихревыми токами в одной из точек на
%% контуре индуктора (для вычисления магнитного потока вихревых токов через
%% поверхность индуктора).
  % Ae1=Aeddy(A, w, sigma, H, R, -z0, a1, mu0)
    %% :: Ae1 = -8.2864e-011 - 6.2722e-011i
    
%% Вычисляемую наводимую вихревыми токами ЭДС индукции в катушке металлоискателя.
  % dE1=-2*i*w*pi*a1*Ae1
    %% :: dE1 = -1.2381e-006 + 1.6357e-006i
    
%% Вычисляем изменение комплексного сопротивления катушки под влиянием объекта
%% (в процессе расчётов считалось, что катушка состоит из одного витка; если 
%% витков n1, результат умножаем на n1*n1).
  % dZ1=-dE1/1
    %% :: dZ1 =  1.2381e-006 - 1.6357e-006i
    
%% Вычисляем соответствующее изменению комплексного сопротивления изменение
%% эквивалентной индуктивности катушки.
  % dL1=imag(dZ1)/w
    %% :: dL1 =   -5.2065e-011